import unittest

import pycodestyle


class TestCodeFormat(unittest.TestCase):

    def test_conformance(self):
        """Test that we conform to PEP-8."""
        style = pycodestyle.StyleGuide()
        result = style.check_files(['ems_r25'])
        self.assertEqual(result.total_errors, 0,
                         "Found code style errors (and warnings).")



import logging

from django.conf import settings
from lxml import etree
from uw_r25 import nsmap, get_resource

from .more_r25 import put_resource


logger = logging.getLogger(__name__)


def update_get_space_ids(ems_rooms):
    """
    Get R25 space_ids for EMS Rooms.

    :param ems_rooms: A collection of ems_client.models.Room
    :return: A dictionary of Room.id: space_id
    """
    space_ids = {}
    for room in ems_rooms:
        if room.active and room.external_reference is not None:
            space_ids[room.id] = room.external_reference

    # while we're here, update the R25 saved search that we'll use
    query_url = "space_search.xml?query_id=%s" % settings.EMS_R25_SPACE_QUERY

    r25_query_tree = get_resource(query_url)

    snode = r25_query_tree.xpath("r25:search", namespaces=nsmap)[0]
    snode.attrib['status'] = 'mod'

    snode = snode.xpath("r25:step", namespaces=nsmap)[0]
    snode.attrib['status'] = 'mod'

    query_modified = False

    found_space_ids = []
    step_param_nbr = -1
    for pnode in snode.xpath("r25:step_param", namespaces=nsmap):
        space_id = pnode.xpath("r25:space_id", namespaces=nsmap)[0].text
        temp = int(pnode.xpath("r25:step_param_nbr", namespaces=nsmap)[0].text)

        found_space_ids.append(space_id)

        if temp > step_param_nbr:
            step_param_nbr = temp

        if space_id not in space_ids.values():
            pnode.attrib['status'] = 'del'
            query_modified = True

    for space_id in space_ids.values():
        if space_id in found_space_ids:
            continue
        step_param_nbr += 1
        param = etree.Element("{%s}step_param" % nsmap['r25'],
                              attrib={'status': 'new'})
        node = etree.Element("{%s}step_param_nbr" % nsmap['r25'])
        node.text = str(step_param_nbr)
        param.append(node)
        node = etree.Element("{%s}space_id" % nsmap['r25'])
        node.text = space_id
        param.append(node)
        snode.append(param)
        query_modified = True

    if query_modified:
        put_resource(query_url, etree.tostring(r25_query_tree))

    return space_ids



import logging
from lxml import etree

from restclients_core.exceptions import DataFailureException
from uw_r25 import nsmap, get_resource
from uw_r25.dao import R25_DAO
from uw_r25.events import events_from_xml
from uw_r25.models import Event


logger = logging.getLogger(__name__)


def live_url(self):
    return (
        "https://25live.collegenet.com/%s/#details&obj_type=event&obj_id=%s" %
        (R25_DAO().get_service_setting('INSTANCE'), self.event_id))


Event.live_url = live_url


class R25ErrorException(Exception):
    """
    This exception means r25 returned <error> elements in a response.

    <?xml version="1.0" encoding="UTF-8"?>
    <r25:results xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                 xmlns:xl="http://www.w3.org/1999/xlink"
                 xmlns:r25="http://www.collegenet.com/r25"
                 pubdate="2019-05-24T00:27:08-07:00"
                 engine="accl">
       <r25:error>
          <r25:msg_id>SY_E_DATAERROR</r25:msg_id>
          <r25:msg>Error saving; data format/validation error</r25:msg>
          <r25:id>789166</r25:id>
       </r25:error>
       <r25:error_details>
          <r25:error_detail table="events"
                            field="event_type_id"
                            value="306"
                            object_id="789166"
                            object_type="4">
            Inactive event_type_id
          </r25:error_detail>
       </r25:error_details>
    </r25:results>
    """
    def __init__(self, **kwargs):
        self.msg_id = kwargs.get('msg_id')
        self.msg = kwargs.get('msg')
        self.object_id = kwargs.get('id')
        self.proc_error = kwargs.get('proc_error')

    def __str__(self):
        return ("Error %s%s: %s%s" % (
            self.msg_id,
            " with %s" % self.object_id if self.object_id else '',
            self.msg,
            ", %s" % self.proc_error if self.proc_error else '',
        ))


class R25MessageException(Exception):
    """
    This exception means r25 returned <messages> elements in a response.

    If the response contains more than one message, the next message is linked
    as 'next_msg'

    <r25:messages>
      <r25:msg_num>1</r25:msg_num>
      <r25:msg_id>EV_I_SPACECON</r25:msg_id>
      <r25:msg_text>Space KNE  225 unavailable due to [rsrv] conflict with
       CENTER FOR HUMAN RIGHTS 10TH ANNIVERSARY [15236046]</r25:msg_text>
      <r25:msg_entity_name>sp_reservations</r25:msg_entity_name>
      <r25:msg_object_id>5326</r25:msg_object_id>
    </r25:messages>
    """
    def __init__(self, num, msg_id, text, entity_name, object_id,
                 next_msg=None):
        self.num = num
        self.msg_id = msg_id
        self.text = text
        self.entity_name = entity_name
        self.object_id = object_id
        self.next_msg = next_msg

    def __str__(self):
        return ("Error %s with %s %s: %s%s" %
                (self.msg_id, self.entity_name, self.object_id, self.text,
                 ' [more...]' if self.next_msg else ''))


class TooManyRequestsException(Exception):
    pass


def post_resource(url):
    """
    Issue a POST request to R25

    :param url: endpoint to POST to
    :return: the response as an lxml.etree
    """

    instance = R25_DAO().get_service_setting('INSTANCE')
    if instance is not None:
        url = "/r25ws/wrd/%s/run/%s" % (instance, url)
    else:
        url = "/r25ws/servlet/wrd/run/%s" % url

    response = R25_DAO().postURL(url, {"Accept": "text/xml"})
    if response.status == 429:
        raise TooManyRequestsException(url)
    if response.status != 201:
        raise DataFailureException(url, response.status, response.data)

    tree = etree.fromstring(response.data.strip())

    # XHTML response is an error response
    xhtml = tree.xpath("//xhtml:html", namespaces=nsmap)
    if len(xhtml):
        raise DataFailureException(url, 500, response.data)

    return tree


def node_as_dict(node):
    mydict = {}
    for element in node:
        name = etree.QName(element).localname
        mydict[name] = element.text

    return mydict


def put_resource(url, body):
    """
    Issue a PUT request to R25

    :param url: endpoint to PUT to
    :param body: text to PUT
    :return: the response as an lxml.etree
    """

    instance = R25_DAO().get_service_setting('INSTANCE')
    if instance is not None:
        url = "/r25ws/wrd/%s/run/%s" % (instance, url)
    else:
        url = "/r25ws/servlet/wrd/run/%s" % url

    headers = {
        "Accept": "text/xml",
        "Content-Type": "text/xml",
    }

    response = R25_DAO().putURL(url, headers, body)
    if response.status not in (200, 201, 400, 403):
        raise DataFailureException(url, response.status, response.data)

    tree = etree.fromstring(response.data.strip())

    # XHTML response is an error response
    xhtml = tree.xpath("//xhtml:html", namespaces=nsmap)
    if len(xhtml):
        raise DataFailureException(url, 500, response.data)

    enodes = tree.xpath("r25:error", namespaces=nsmap)
    if len(enodes):
        err = node_as_dict(enodes[0])
        raise R25ErrorException(**err)

    mnodes = tree.xpath("r25:messages", namespaces=nsmap)
    if len(mnodes):
        next_ex = None
        for mnode in reversed(mnodes):
            next_ex = R25MessageException(
                mnode.xpath("r25:msg_num", namespaces=nsmap)[0].text,
                mnode.xpath("r25:msg_id", namespaces=nsmap)[0].text,
                mnode.xpath("r25:msg_text", namespaces=nsmap)[0].text,
                mnode.xpath("r25:msg_entity_name", namespaces=nsmap)[0].text,
                mnode.xpath("r25:msg_object_id", namespaces=nsmap)[0].text,
                next_ex)
        raise next_ex

    return tree


def delete_resource(url):
    """
    Issue a DELETE request to R25

    :param url: endpoint to DELETE
    :return: the response as an lxml.etree
    """

    instance = R25_DAO().get_service_setting('INSTANCE')
    if instance is not None:
        url = "/r25ws/wrd/%s/run/%s" % (instance, url)
    else:
        url = "/r25ws/servlet/wrd/run/%s" % url

    headers = {
        "Accept": "text/xml",
        "Content-Type": "text/xml",
    }

    response = R25_DAO().deleteURL(url, headers)
    if response.status != 200:
        raise DataFailureException(url, response.status, response.data)

    tree = etree.fromstring(response.data.strip())

    # XHTML response is an error response
    xhtml = tree.xpath("//xhtml:html", namespaces=nsmap)
    if len(xhtml):
        raise DataFailureException(url, 500, response.data)

    return tree


def update_value(node, name, value):
    """
    Adds or updates the value of a basic text element in an R25 etree.

    Adds 'status="mod"' to all ancestor elements so that R25 recognizes our
    change.

    :param node: The node which contains our element
    :param name: The element's name
    :param value: The new value for the element
    :return: The new or updated element
    """

    try:
        element = node.xpath("r25:%s" % name, namespaces=nsmap)[0]
    except IndexError:
        # create the element
        element = etree.SubElement(node, "{%s}%s" % (nsmap['r25'], name),
                                   nsmap=nsmap)

    if (not value and not element.text) or element.text == value:
        # no change
        return element

    logger.debug("changing %s from %s to %s" %
                 (element.getroottree().getpath(element), element.text, value))
    element.text = value

    # mark ancestors as modified
    while 'status' in node.attrib and node.attrib['status'] == 'est':
        node.attrib['status'] = 'mod'
        node = node.getparent()

    return element


def add_node(node, name):
    """
    Adds a new node to an R25 etree

    Adds 'status="mod"' to all ancestor elements so that R25 recognizes our
    change.

    :param node: The parent of our new node
    :param name: The new node's name
    :return: The new node
    """

    logger.debug("adding %s to %s" % (name, node.getroottree().getpath(node)))

    element = etree.SubElement(node, "{%s}%s" % (nsmap['r25'], name),
                               attrib={'status': 'new'}, nsmap=nsmap)

    # mark ancestors as modified
    while 'status' in node.attrib and node.attrib['status'] == 'est':
        node.attrib['status'] = 'mod'
        node = node.getparent()

    return element


def delete_node(node):
    """
    Deletes a node from an R25 etree

    Just marks the node with 'status="del"' and adds 'status="mod"' to all
    ancestor elements so that R25 recognizes our change.

    :param node: The node to mark deleted
    :return: The node
    """

    logger.debug("deleting %s" % node.getroottree().getpath(node))

    node.attrib['status'] = 'del'
    node = node.getparent()

    # mark ancestors as modified
    while 'status' in node.attrib and node.attrib['status'] == 'est':
        node.attrib['status'] = 'mod'
        node = node.getparent()

    return node


def update_event(event):
    """
    Create or update the given event in R25

    Only the features supported by uw_r25 are supported here.
    Some unsupported fields are initialized with hard-coded values suitable for
    events from ASA's EMS.

    :param event: a uw_r25.models.event
    :return: the new or updated event from R25, as a uw_r25.models.event
    """

    # Start with xml from R25, by requesting a new blank event, or getting the
    # editable version of our existing event
    if event.event_id is None:
        # Create a new event
        url = "events.xml"
        event_tree = post_resource(url)
        enode = event_tree.xpath("r25:event", namespaces=nsmap)[0]

        event.event_id = enode.xpath("r25:event_id", namespaces=nsmap)[0].text

        # initialize some things that aren't kept in the uw_r25 model
        update_value(enode, 'node_type', 'E')
        update_value(enode, 'event_type_id',
                     R25_DAO().get_service_setting('EVENTTYPE'))

        onode = enode.xpath("r25:organization", namespaces=nsmap)[0]
        update_value(onode, 'organization_id',
                     R25_DAO().get_service_setting('ORGANIZATION'))
        update_value(onode, 'primary', 'T')

        # delete the blank profile
        pnode = enode.xpath("r25:profile", namespaces=nsmap)[0]
        enode.remove(pnode)

    else:
        url = "event.xml?event_id=%s&mode=edit" % event.event_id
        event_tree = get_resource(url)
        enode = event_tree.xpath("r25:event", namespaces=nsmap)[0]

    update_value(enode, 'alien_uid', event.alien_uid)
    update_value(enode, 'event_name', event.name)
    update_value(enode, 'event_title', event.title)
    update_value(enode, 'start_date', event.start_date)
    update_value(enode, 'end_date', event.end_date)
    update_value(enode, 'state', event.state)
    update_value(enode, 'parent_id', event.parent_id)
    update_value(enode, 'cabinet_id', event.cabinet_id)
    update_value(enode, 'cabinet_name', event.cabinet_name)

    # add or update each reservation/profile
    # only one reservation per profile is supported
    for res in event.reservations:
        if res.reservation_id:
            # find existing profile
            xpath = \
                "r25:profile[./r25:reservation/r25:reservation_id = '%s']" % \
                res.reservation_id
            pnode = enode.xpath(xpath, namespaces=nsmap)[0]
            rnode = pnode.xpath("r25:reservation", namespaces=nsmap)[0]

        else:
            # add new profile and reservation
            pnode = add_node(enode, 'profile')
            rnode = add_node(pnode, 'reservation')

        update_value(pnode, 'profile_name', res.profile_name)
        update_value(pnode, 'init_start_dt', res.start_datetime)
        update_value(pnode, 'init_end_dt', res.end_datetime)

        update_value(rnode, 'reservation_start_dt', res.start_datetime)
        update_value(rnode, 'reservation_end_dt', res.end_datetime)
        update_value(rnode, 'reservation_state', res.state)

        # add or update space_reservation
        # only one space_reservation per reservation is supported
        srnode = None
        try:
            srnode = rnode.xpath("r25:space_reservation", namespaces=nsmap)[0]
        except IndexError:
            pass

        if res.space_reservation is not None:
            if srnode is None:
                # Add space reservation
                srnode = add_node(rnode, 'space_reservation')

            update_value(srnode, 'space_id', res.space_reservation.space_id)
        elif srnode is not None:
            # outdated space reservation. delete it
            delete_node(srnode)

    # Make sure event dates encompass all reservations
    # for res in r25_event.reservations:
    #     res_start_date = res.start_datetime.split('T')[0]
    #     res_end_date = res.end_datetime.split('T')[0]
    #     if res_start_date < r25_event.start_date:
    #         r25_event.start_date = res_start_date.isoformat()
    #     if res_end_date > r25_event.end_date:
    #         r25_event.end_date = res_end_date

    if enode.attrib['status'] == 'est':
        logger.debug("Event unchanged")
        return event

    url = "event.xml?event_id=%s" % event.event_id

    return events_from_xml(put_resource(url, etree.tostring(event_tree)))[0]


def delete_event(event_id):
    """
    Delete event from R25

    :param event_id: an R25 event id
    :return: the response from R25, as an lxml.etree
    """

    url = "event.xml?event_id=%s" % event_id

    result = delete_resource(url)

    return result



import datetime
import logging
import re

from dateutil.parser import parse
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from ems_client.models import Status
from ems_client.service import Service
from restclients_core.exceptions import DataFailureException
from urllib3.exceptions import HTTPError
from uw_r25.events import get_event_by_alien_id, get_event_by_id
from uw_r25.models import Event, Reservation, Space

from ems_r25.more_r25 import (delete_event, update_event,
                              R25MessageException, R25ErrorException,
                              TooManyRequestsException
                              )
from ems_r25.utils import update_get_space_ids


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'adds or updates R25 events with events from EMS'

    def add_arguments(self, parser):
        parser.add_argument(
            '-s',
            '--start',
            help="Start of date range. Default is today, or yesterday if "
                 "using --changed.",
        )
        parser.add_argument(
            '-e',
            '--end',
            help="End of date range. Default is START+7 days, or today if "
                 "using --changed.",
        )

        parser.add_argument(
            '-c',
            '--changed',
            action='store_true',
            help="Get Bookings that have changed within date range. Includes "
                 "Bookings where the Reservation has changed, even if the "
                 "Booking has not.",
        )

        parser.add_argument(
            '-d',
            '--delete',
            action='store_true',
            help='Delete matched R25 Events',
        )

        parser.add_argument(
            '-u',
            '--update',
            action='store_true',
            help='Update R25 Events',
        )

    def handle(self, *args, **options):
        if options['changed']:
            if options['start']:
                start_date = parse(options['start']).date()
            else:
                start_date = datetime.date.today()
            if options['end']:
                end_date = parse(options['end']).date()
            else:
                end_date = datetime.date.today()

        else:
            if options['start']:
                start_date = parse(options['start']).date()
            else:
                start_date = datetime.date.today()
            if options['end'] == 'max':
                end_date = datetime.date.max
            elif options['end']:
                end_date = parse(options['end']).date()
            else:
                end_date = start_date + datetime.timedelta(days=7)

        _ems = Service()

        try:
            space_ids = update_get_space_ids(_ems.get_all_rooms())
        except R25ErrorException as ex:
            raise CommandError("Unable to update space search: %s" % ex)

        status_list = _ems.get_statuses()
        statuses = {}
        search_statuses = []
        for status in status_list:
            statuses[status.id] = status
            if status.description not in settings.EMS_R25_IGNORE_STATUSES:
                search_statuses.append(status.id)

        # Get all bookings in range, regardless of room, status, or event type.
        # We do this because a now-unwanted booking might already have been
        # Created in R25, and we need to cancel it there.
        bookings = _ems.get_changed_bookings(
            start_date=start_date.isoformat(), end_date=end_date.isoformat(),
            statuses=search_statuses
        ) if options['changed'] else _ems.get_bookings(
            start_date=start_date.isoformat(), end_date=end_date.isoformat(),
            statuses=search_statuses
        )

        ems_reservations = {}
        for booking in bookings:
            if options['changed']:
                # Booking hasn't changed, EMS Reservation has...
                if (booking.date_changed.date() < start_date or
                        booking.date_changed.date() > end_date):
                    continue

            booking.status = statuses[booking.status_id]

            if booking.reservation_id not in ems_reservations:
                # Use data from first booking as reservation data.
                # FIXME: we need a way to grab actual EMS Reservation data
                # FIXME: instead of just Bookings
                ems_reservations[booking.reservation_id] = booking
                ems_reservations[booking.reservation_id].bookings = {}
            ems_reservations[
                booking.reservation_id].bookings[booking.id] = booking

        for ems_res_id in ems_reservations:
            ems_reservation = ems_reservations[ems_res_id]
            logger.info("Processing EMS Reservation %d" % ems_res_id)

            # alien_uid is how we tie EMS Reservation to R25 Event.
            r25_alien_uid = "AT_EMS_RSRV_%s" % ems_res_id

            r25_event = None
            try:
                r25_event = get_event_by_alien_id(r25_alien_uid)
                logger.debug("\tFound R25 event %s: '%s'" %
                             (r25_event.event_id, r25_event.name))

            except DataFailureException:
                # No R25 event matching this EMS event
                pass
            except HTTPError as ex:
                # Server timeout, etc
                self.stdout.write("HTTP Error retrieving R25 Event, skipping "
                                  "Reservation %s: %s" % (ems_res_id, ex))
                continue

            if options['delete']:
                if r25_event:
                    logger.debug("\tDeleting!")
                    delete_event(r25_event.event_id)
                else:
                    logger.debug("\tNothing to delete.")
                continue

            if r25_event is None:
                r25_event = Event()
                r25_event.alien_uid = r25_alien_uid
                r25_event.reservations = []

            r25_event.name = ems_reservation.event_name[:40].strip()
            r25_event.title = ems_reservation.event_name.strip()
            r25_event.state = r25_event.CONFIRMED_STATE

            ems_bookings = ems_reservation.bookings
            for ems_bk_id in ems_bookings:
                ems_booking = ems_bookings[ems_bk_id]

                logger.debug("\tProcessing EMS Booking %d: '%s'" %
                             (ems_bk_id, ems_booking.event_name))
                logger.debug("\t\tStatus: %s, space_id: %s" % (
                             (ems_booking.status.description,
                              space_ids.get(ems_booking.room_id))))
                logger.debug("\t\tStart: %s, End: %s, Changed: %s" % (
                    ems_booking.time_booking_start.isoformat(),
                    ems_booking.time_booking_end.isoformat(),
                    ems_booking.date_changed.isoformat()))

                # profile_name is how we tie EMS Booking to R25 Reservation.
                # We only use the most basic type of profile, so profile to
                # reservation is a 1:1 relationship.
                r25_profile_name = "AT_EMS_BOOK_%s" % ems_bk_id

                r25_res = None
                for r in r25_event.reservations:
                    if r.profile_name == r25_profile_name:
                        logger.debug(
                            "\t\tFound R25 reservation/profile %s: '%s'" %
                            (r.reservation_id, r.profile_name))
                        logger.debug(
                            "\t\t\tState: %s, Space Reservation: %s" %
                            (r.state_name(),
                             r.space_reservation.space_id if
                             r.space_reservation else ''))
                        r25_res = r
                        break

                # does this Booking belong in R25?
                if (ems_booking.status_type_id ==
                        Status.STATUS_TYPE_BOOKED_SPACE and
                        ems_booking.room_id in space_ids and
                        ems_booking.status.description not in
                        settings.EMS_R25_REMOVE_STATUSES):

                    # We want a reservation. Create if necessary.
                    if r25_res is None:
                        r25_res = Reservation()
                        r25_res.profile_name = r25_profile_name
                        r25_res.space_reservation = None
                        r25_event.reservations.append(r25_res)

                    r25_res.start_datetime = \
                        ems_booking.time_booking_start.isoformat()
                    r25_res.end_datetime = \
                        ems_booking.time_booking_end.isoformat()
                    r25_res.state = r25_res.STANDARD_STATE
                    if r25_res.space_reservation is None:
                        r25_res.space_reservation = Space()

                    r25_res.space_reservation.space_id = space_ids[
                        ems_booking.room_id]

                # does not belong. Does one already exist?
                elif r25_res is not None:
                    # Cancel this unwanted r25 reservation
                    r25_res.state = r25_res.CANCELLED_STATE
                    r25_res.space_reservation = None

            # if this R25 Event is empty, there's nothing to do
            if not r25_event.reservations:
                logger.debug("\tNo existing R25 reservations and no wanted "
                             "new R25 reservations")
                continue

            # if there are no active reservations left, cancel the event
            has_active_reservations = False
            for r in r25_event.reservations:
                if r.state != r.CANCELLED_STATE:
                    has_active_reservations = True
                    break
            if not has_active_reservations:
                r25_event.state = r25_event.CANCELLED_STATE

            # by default, don't actually make changes
            if not options['update']:
                continue

            try:
                logger.debug("\tUpdating event")
                updated = update_event(r25_event)

            except R25MessageException as ex:
                while ex:
                    if ex.msg_id == 'EV_I_SPACECON':
                        logger.warning(
                            "Conflict while syncing EMS Reservation %s: %s" %
                            (ems_reservation.reservation_id, ex.text))
                        match = re.search(r'\[(?P<event_id>\d+)\]', ex.text)
                        old_event = get_event_by_id(match.group('event_id'))
                        logger.warning(
                            "Existing event: %s" % old_event.live_url())
                        logger.warning(
                            "Is blocking event: %s" % r25_event.live_url())

                    else:
                        logger.warning(
                            "R25 message while syncing EMS Reservation %s to "
                            "R25 Event %s: %s" % (
                                ems_reservation.reservation_id,
                                r25_event.event_id, ex))

                    ex = ex.next_msg

            except R25ErrorException as ex:
                logger.warning(
                    "R25 error while syncing EMS Reservation %s to R25 Event "
                    " %s: %s" % (ems_reservation.reservation_id,
                                 r25_event.event_id, ex))

            except HTTPError as ex:
                logger.warning(
                    "HTTP error while syncing EMS Reservation %s to R25 Event "
                    " %s: %s" % (ems_reservation.reservation_id,
                                 r25_event.event_id, ex))

            except TooManyRequestsException:
                self.stdout.write(
                    "Too Many Requests while syncing EMS Reservation %s to "
                    "R25 Event %s" % (ems_reservation.reservation_id,
                                      r25_event.event_id))
